#include <lektor/common.h>
#include <lektor/net.h>
#include <lektor/lib/strv.h>

PRIVATE_FUNCTION bool
___crop_string(char *string, const size_t max_len)
{
    const size_t len       = strlen(string);
    const bool was_changed = max_len < len;

    /* Here we set all the string to '\0' because we don't want to handle the rest of the command */
    for (size_t i = max_len; i < len; ++i)
        string[i] = '\0';

    return was_changed;
}

void
lkt_command_log_to_debug(const struct lkt_command *cmd)
{
    LOG_DEBUG("CMD-DBG", "name: %s", cmd->name);
    for (size_t i = 0; cmd->args[i]; ++i)
        LOG_DEBUG("CMD-DBG", "arg[%zu] = %s", i, cmd->args[i]);
    LOG_WARN_IF(cmd->was_truncated, "CMD-DBG", "Command named '%s' was truncated!", cmd->name);
}

struct lkt_command
lkt_command_parse(char *raw)
{
    errno                  = 0;
    char *endptr           = NULL;
    bool skip_cont         = false;
    struct lkt_command res = {
        .name          = raw,
        .args          = { 0 },
        .was_truncated = false,
        .cont          = (raw != NULL) ? strtol(raw, &endptr, 10) : 0,
    };

    /* There was a continuation in the command */
    if ((errno != 0) || (endptr == raw)) {
        res.cont  = 0;
        skip_cont = true;
    }

    unless (skip_cont) {
        RETURN_IF(*endptr == '\0', "Error in strtol", res);
        raw += strspn(raw, " 1234567890");
        RETURN_IF(*raw == '\0', "No command name found!", res);
        res.name = raw;
    }

    /* From here, `raw` is a string, the name and the args are comma separated.
     * If no argument is present we only have the command name! */
    raw += strcspn(raw, " ");
    if (*raw == '\0')
        return res;

    /* Found a space, there are arguments. The `raw` pointer will points to the
     * begening of the first argument. */
    FAIL_UNLESS(raw[0] == ' ', "Must be a space here");
    raw[0] = '\0';
    raw++;

#define ___END_OF_ARGUMENTS_FOUND                                                               \
    {                                                                                           \
        LOG_DEBUG("CMD", "Found early stop in arguments for command %s, argument count is %zu", \
                  res.name, i + 1);                                                             \
        break;                                                                                  \
    }
#define ___SKIP_OVER_CHAR(___char)                                     \
    {                                                                  \
        FAIL_UNLESS(raw[0] == ___char, "Must be a " #___char " here"); \
        *raw = '\0';                                                   \
        raw += 1;                                                      \
    }

    /* Find all the arguments */
    for (size_t i = 0; (i < LKT_MESSAGE_ARGS_MAX) && (*raw != '\0'); ++i) {
        while (strv_isspace(*raw))
            raw++;

        if (*raw == '"') {
            raw += 1;
            res.args[i] = raw;
            raw += strcspn(raw, "\"");
            if (*raw != '\0')
                ___SKIP_OVER_CHAR('"')
            else
                ___END_OF_ARGUMENTS_FOUND
        }

        else if (*raw == '\'') {
            raw += 1;
            res.args[i] = raw;
            raw += strcspn(raw, "'");
            if (*raw != '\0')
                ___SKIP_OVER_CHAR('\'')
            else
                ___END_OF_ARGUMENTS_FOUND
        }

        else {
            res.args[i] = raw;
            raw += strcspn(raw, " ");
            if (*raw != '\0')
                ___SKIP_OVER_CHAR(' ')
            else
                ___END_OF_ARGUMENTS_FOUND
        }
    }

#undef ___END_OF_ARGUMENTS_FOUND
#undef ___SKIP_OVER_CHAR

    return res;
}

struct lkt_command
lkt_command_clone(struct lkt_command *c)
{
    struct lkt_command res = {
        .name = strdup(c->name),
        .args = { 0 },
    };

    for (size_t i = 0; c->args[i] && i < LKT_MESSAGE_ARGS_MAX; i++)
        res.args[i] = strdup(c->args[i]);

    return res;
}

void
lkt_command_free(struct lkt_command *c)
{
    if (c->name) {
        free(c->name);
        c->name = NULL;
    } else {
        LOG_ERROR("CMD", "Double free for a command! This is an error");
    }

    for (size_t i = 0; c->args[i]; i++) {
        free(c->args[i]);
        c->args[i] = NULL;
    }
}
